#define __NEWICKFORM_C__

#include "seqUtil.h"
#include "Newickform.h"
#include <math.h>
#include <stdlib.h>

void MakeTipsZero(newick_node *root)
{	
	newick_child *child;
	if (root->childNum == 0)
	{
		root->longestbranch=0;
	}
	else
	{
		child = root->child;
		while (child != NULL)
		{
			MakeTipsZero(child->node);
			child = child->next;
		}
		
	}
}

void FindNumElements(newick_node *root)
{	
	newick_child *child;
	root->numElements=0;
	if (root->childNum == 0)
	{
		root->numElements=1;
	}
	else
	{
		child = root->child;
		while (child != NULL)
		{
			child->node->parent=root;
			FindNumElements(child->node);
			if (child->node->cluster!=1)
			{
				root->numElements=root->numElements+child->node->numElements;
			}
			child = child->next;
		}		
	}
}
int AverageBranch(newick_node *root)
{	
	newick_child *child;
	float totalDistance;
	float averagetotal=0;
	float longesttotal=0;
	float total=0;
	int times=0;
	int longesttime=0;
	
	child = root->child;
	
	while (child != NULL)
	{
		times++;
		
		if (child->node->cluster!=1)
		{
			
			if (child->node->childNum==0) 
			{
				totalDistance=child->node->dist;
				child->node->totaldistance=0;
			}
			else
			{
				AverageBranch(child->node);
				totalDistance=child->node->numElements*((child->node->dist) + (child->node->totaldistance));
			}
			
			
			if (totalDistance>=longesttotal) 
			{
				longesttotal=totalDistance;
				longesttime=times;
			}
			total=total+(totalDistance);
		}
		child = child->next;
	}
	
	if (root->numElements!=0)
	{
		averagetotal=total/root->numElements;
	}
	else
	{
		averagetotal=0;
	}
	
	root->totaldistance=averagetotal;
	return longesttime;
}

float rerootOption(newick_node *root, float totaldistance, int rerooted)
{
	newick_child *child;
	child = root->child;
	float rerootoption=totaldistance;
	int childrank=0;
	while (child != NULL)
	{
		if (child->node->cluster!=1)
		{
			childrank++;
			child->node->cluster=1;
			root->numElements=root->numElements-child->node->numElements;
			AverageBranch(root);
			if (rerooted==1)
			{
				child->node->totaldistance=rerootoption;
			}
			float newrerootoption=root->numElements*(root->totaldistance + child->node->dist) + child->node->numElements*(child->node->totaldistance);
			newrerootoption=newrerootoption/(root->numElements+child->node->numElements);
			
			if (newrerootoption<rerootoption)
			{
				child->node->roothere=1;
				root->roothere=0;
				rerootoption=newrerootoption;
			}
			else
			{
				child->node->roothere=0;
			}
			rerooted=1;
			float rerootoption1;
			rerootoption1=rerootoption;
			rerootoption=rerootOption(child->node,rerootoption,rerooted);
			if ((rerootoption1>=rerootoption) && childrank!=1)
			{
				child->node->roothere=0;
			}
			child->node->cluster=0;
			AverageBranch(child->node);
			root->numElements=root->numElements+child->node->numElements;
		}
		child=child->next;
	}
	
	return rerootoption;
}



int CutAverageBranch(newick_node *root, int maxMutation, fReturn *farray, float L, int timepoint, int timepointcluster)
{
	newick_child *child;
	float MaxMut=maxMutation;
	int longesttime=AverageBranch(root);
	if (L*(root->totaldistance)<MaxMut)
	{
		root->possibleroothere=1;
	}
	while (L*(root->totaldistance)>=MaxMut)
	{
		float totaldistance=root->totaldistance;
		float rerootoption=rerootOption(root,totaldistance,0);
		
		if (L*rerootoption<MaxMut)
		{
			break;
		}
		
		longesttime=AverageBranch(root);
		
		
		int o;
		child=root->child;
		for (o=1;o<longesttime;o++)
		{
			child = child->next;
		}
		
		child->node->cluster=1;
		(farray->farray[child->node->numElements])++;
		if (child->node->numElements!=0)
		{
			printTaxon(child->node, timepoint, timepointcluster);
			timepointcluster++;
		}
		newick_child *child2;
		child2=child->node->child;
		child->node->clusterless=1;
		while (child2 != NULL)
		{
			if (child2->node->roothere!=1 && child2->node->cluster!=1)
			{
				child->node->clusterless=0;
			}
			child2 = child2->next;
		}
		if (child->node->possibleroothere==1 || child->node->childNum==0 || (child->node->numElements==1 && child->node->clusterless!=1))
		{
			child->node->roothere=1;
			
		}

		root->numElements=root->numElements - child->node->numElements;
		AverageBranch(root);
		if (L*(root->totaldistance)<MaxMut)
		{
			root->possibleroothere=1;
		}
		
	}
	return timepointcluster;

}


void SharedBranch(newick_node *root)
{	
	newick_child *child;
	float branchDistance;
	float longestdistance=0;
	
	child = root->child;
	
	while (child != NULL)
	{
		if (child->node->cluster!=1)
		{
			if (child->node->childNum==0) 
			{
				branchDistance=0;
			}
			else
			{
				branchDistance=(child->node->dist) + (child->node->longestbranch);
			}
			if (branchDistance >= longestdistance)
			{
				longestdistance=branchDistance;
			}
			
		}
		
		child = child->next;
	}
	
	root->longestbranch=longestdistance;
	
}




int CountClusters(newick_node *root, fReturn *farray, int Threshold, int L, int maxMutation, int timepoint, int timepointcluster)
{
	newick_child *child;	
	if (root->childNum != 0)
	{		
		child = root->child;
		while (child != NULL && child->node->cluster!=1)
		{
			timepointcluster=CountClusters(child->node, farray, Threshold, L, maxMutation, timepoint, timepointcluster);
			child = child->next;
		}
		
		FindNumElements(root);
		
		timepointcluster=CutAverageBranch(root, maxMutation, farray, L, timepoint, timepointcluster);
		
		SharedBranch(root);
		
		if (rint(L*(root->longestbranch + root->dist))>Threshold)
		{
			root->cluster=1;
			(farray->farray[root->numElements])++;
			if (root->numElements!=0)
			{
				printTaxon(root, timepoint, timepointcluster);
				timepointcluster++;
			}
			if (root->parent!=NULL)
			{
				root->parent->numElements=root->parent->numElements - root->numElements;			
			}
			
			root->clusterless=1;
			child = root->child;
			while (child != NULL)
			{
				if (child->node->roothere!=1 && child->node->cluster!=1)
				{
					root->clusterless=0;
				}
				child = child->next;
			}
			
			if ((root->possibleroothere==1 || root->childNum==0 || (root->numElements==1 && root->clusterless!=1)) && root->numElements>0)
			{
				root->roothere=1;
			}

		}
		
		
	}
	return timepointcluster;

	
}

void printRepeatTree(newick_node *root, FILE *g, int numbranch, int* multiple, char ** constantarray)
{	
	newick_child *child;
	int w;
	int v;
	if (root->childNum == 0)
	{
		for (w=0; w<numbranch; w++)
		{
			if (strcmp (root->taxon, constantarray[w])==0)
			{	
				if (multiple[w]>1)
				{
					for (v=1; v<multiple[w]; v++)
					{
						fprintf(g,"%s:%f,", root->taxon, root->dist);
					}
					fprintf(g,"%s:%f", root->taxon, root->dist);
				}
				else if (multiple[w]==1)
				{
					fprintf(g,"%s:%f", root->taxon, root->dist);
				}
				else if (multiple[w]==0)
				{
					fprintf(g,"recombinant:0.0");
				}
				
			}
		}
	}
	else
	{
		child = root->child;
		if (root->dist != 0.0)
		{
			fprintf(g,"(");
		}
		while (child != NULL)
		{
			printRepeatTree(child->node, g, numbranch, multiple, constantarray);
			if (child->next != NULL)
			{
				fprintf(g,",");
			}
			child = child->next;
		}
		if (root->dist != 0.0) {
			if (root->taxon != NULL)
			{
				fprintf(g,")%s:%f", root->taxon, root->dist);
			}
			else
			{
				fprintf(g,"):%f", root->dist);
			}
		}
		
	}
}






int printTaxon(newick_node * root, int timepoint, int timepointcluster)
{
	root->timepoint=timepoint;
	newick_child *child;
	if (root->childNum == 0)
	{
		
		if (root->recombinant!=1 && root->printed!=1)
		{
			//printf("%s, ", root->taxon);

		}
		root->printed=1;
		
	}
	else
	{
		child = root->child;
		
		while (child != NULL)
		{
			
			timepointcluster=printTaxon(child->node, timepoint, timepointcluster);
			
			
			child = child->next;
		}
		
	}
	return timepointcluster;
}


int printTaxonNew(newick_node * root, int roothere, int numroots, int timepoint, FILE *j)
{
	newick_child *child;
	if (root->childNum == 0)
	{
		if (root->recombinant!=1 && root->printed!=1)
		{
			
			roothere=1;
			int r;
			for (r=0; r<5; r++)
			{
				if (r!=timepoint)
				{
					root->clusternumbers[r]=0;
				}
				else
				{
					root->clusternumbers[r]=numroots;
				}
			}
			fprintf(j,"%s, ", root->taxon);
			
		}
		root->printed=1;
		
		
	}
	else
	{
		child = root->child;
		
		while (child != NULL)
		{
			
			roothere=printTaxonNew(child->node, roothere, numroots, timepoint, j);
			
			child = child->next;
		}
		
	}
	return roothere;
}

void makePrintedZero(newick_node * root)
{
	newick_child *child;
	if (root->childNum == 0)
	{
		
		root->printed=0;
		
	}
	else
	{
		child = root->child;
		
		while (child != NULL)
		{
			
			makePrintedZero(child->node);
			
			child = child->next;
		}
		
	}
}


newick_node* parseTree(const char *const str)
{
	newick_node *const node = seqMalloc(sizeof *node);
	const char *pcCurrent;
	const char *pcStart = str;
	const char *pcColon = NULL;
	
	if (sDEBUG == 1)
	{
		printf("%s\n", str);
	}
	
	if (*pcStart != '(')
	{
		// Leaf node. Separate taxon name from distance. If distance not exist then take care of taxon name only
		pcCurrent = str;
		while (*pcCurrent != '\0' && *pcCurrent != ')' && *pcCurrent != ',')
		{
			if (*pcCurrent == ':')
			{
				pcColon = pcCurrent;
			}
			pcCurrent++;
		}
		if (pcColon == NULL)
		{
			// Taxon only
			node->taxon = seqMalloc(strlen(pcStart) + 1);
			memcpy(node->taxon, pcStart, strlen(pcStart));
			node->dist = -1;
		}
		else
		{
			// Taxon
			node->taxon = seqMalloc(pcColon - pcStart + 1);
			memcpy(node->taxon, pcStart, pcColon - pcStart);
			// Distance
			pcColon++;
			node->dist = atof(pcColon);
		}
		node->childNum = 0;
	}
	else
	{
		newick_child *child = NULL;
		// Search for all child nodes
		// Find all ',' until corresponding ')' is encountered
		int iCount = 0;
		pcStart++;
		pcCurrent = pcStart;
		while (iCount >= 0)
		{
			switch (*pcCurrent)
			{
				case '(':
					// Find corresponding ')' by counting
					pcStart = pcCurrent;
					pcCurrent++;
					iCount++;
					while (iCount > 0)
					{
						if (*pcCurrent == '(')
						{
							iCount++;
						}
						else if (*pcCurrent == ')')
						{
							iCount--;
						}
						else if (*pcCurrent == '\0') 
						{
						    fprintf(stderr,"Unterminated parens\n");
						    exit(101);
						}
						pcCurrent++;
					}
					while (*pcCurrent != ',' && *pcCurrent != ')')
					{
						if (*pcCurrent == '\0') 
						{ 
						    fprintf(stderr,"Unterminated parens\n");
						    exit(102);
						}   
						pcCurrent++;
					}
 					//cTemp = *pcCurrent;
					//*pcCurrent = '\0';
					// Create a child node
					if (child == NULL)
					{
						node->child = seqMalloc(sizeof *node->child);
						node->childNum = 1;
						child = node->child;
					}
					else
					{
						child->next = seqMalloc(sizeof *child->next);
						node->childNum++;
						child = child->next;
					}
					child->node = parseTree(pcStart);
					child->node->parent=node;
					//*pcCurrent = cTemp;
					if (*pcCurrent != ')')
					{
						pcCurrent++;
					}
					break;
					
				case ')':
					// End of this tree. Go to next part to retrieve distance
					iCount--;
					node->dist=-1;
					break;
					
				case ',':
					// Impossible separation since according to the algorithm, this symbol will never encountered.
					// Currently don't handle this and don't create any node
					fprintf(stderr,"Extra comma?\n");
					exit(103);
					break;
					
				default:
					// leaf node encountered
					pcStart = pcCurrent;
					while (*pcCurrent != ',' && *pcCurrent != ')')
					{
						if (*pcCurrent == '\0') 
						{ 
						    fprintf(stderr,"Unterminated parens\n");
						    exit(104);
						}   
						pcCurrent++;
					}
					// cTemp = *pcCurrent;
					// *pcCurrent = '\0';
					// Create a child node
					if (child == NULL)
					{
						node->child = seqMalloc(sizeof *node->child);
						node->childNum = 1;
						child = node->child;
					}
					else
					{
						child->next = seqMalloc(sizeof *child->next);
						node->childNum++;
						child = child->next;
					}
					child->node = parseTree(pcStart);
					child->node->parent=node;
					// *pcCurrent = cTemp;
					if (*pcCurrent != ')')
					{
						pcCurrent++;
					}
					break;
			}
		}
		
		// If start at ':', then the internal node has no name.
		
		pcCurrent++;
		
		if (*pcCurrent == ':')
		{
			pcStart = pcCurrent + 1;
			while (*pcCurrent != '\0' && *pcCurrent != ';' && *pcCurrent != ')'  && *pcCurrent != '(' && *pcCurrent != ',')
			{
				pcCurrent++;
			}
			// cTemp = *pcCurrent;
			// *pcCurrent = '\0';
			node->dist = atof(pcStart);
			// *pcCurrent = cTemp;
		}
		
		else if (*pcCurrent != ';' && *pcCurrent != '\0')
		{
			// Find ':' to retrieve distance, if any.
			// At this time *pcCurrent should equal to ')'
			
			pcStart = pcCurrent;
			
			while ((*pcCurrent != ',') && (*pcCurrent != ':') && (*pcCurrent != '\0') && (*pcCurrent != ')')  && (*pcCurrent != '(') && (*pcCurrent != ';'))
			{
				pcCurrent++;
				
			}
			// cTemp = *pcCurrent;
			// *pcCurrent = '\0';
			node->taxon = seqMalloc(pcCurrent - pcStart + 1);
			memcpy(node->taxon, pcStart, pcCurrent - pcStart);
			// *pcCurrent = cTemp;
			
			if (*pcCurrent == ':')
			{
				pcCurrent++;
				
				pcStart = pcCurrent;
				while (*pcCurrent != '\0' && *pcCurrent != ';' && *pcCurrent != ')' && *pcCurrent != ':' && *pcCurrent != '(' && *pcCurrent != ',')
				{
					pcCurrent++;
				}
				// cTemp = *pcCurrent;
				// *pcCurrent = '\0';
				
				node->dist = atof(pcStart);
				// *pcCurrent = cTemp;
			}
			
			
		}
		
	}
	return node;
}




void printTree(newick_node *root, FILE *g)
{	
	newick_child *child;
	if (root->childNum == 0)
	{
		fprintf(g,"%s:%f", root->taxon, root->dist);
	}
	else
	{
		child = root->child;
		if (root->dist != 0.0)
		{
			fprintf(g,"(");
		}
		while (child != NULL)
		{
			printTree(child->node, g);
			if (child->next != NULL)
			{
				fprintf(g,",");
			}
			child = child->next;
		}
		if (root->dist != 0.0) {
			if (root->taxon != NULL)
			{
				fprintf(g,")%s:%f", root->taxon, root->dist);
			}
			else
			{
				fprintf(g,"):%f", root->dist);
			}
		}
		
	}
}





newick_node* changeRoot(newick_node* root, newick_child* newn, rerootReturn* rerootbranch)
{
	if (rerootbranch->correspondingChild == NULL)
	{
		float tmp;
		tmp=(root->dist);
		(root->dist)=(root->child->node->dist);
		(root->child->node->dist)=tmp;
		(root->childNum)--;
		(root->child->node->childNum)++;
		
		newick_node *Atmp;
		newick_node *Btmp;
		newick_child *Ctmp;
		newick_child *Dtmp;
		
		
		Atmp=root;
		Btmp=root->child->node;
		Dtmp=root->child;
		Ctmp=root->child->node->child;
		
		root=root->child->node;
		newn->node=Atmp;
		newn->next=Ctmp;
		root->child=newn;
		
		(root->child->node->child)=(root->child->node->child->next);
		
		root->parent=NULL;
		
		
		
		
		
	}
	
	else
	{
		
		newick_node *Atmp;
		newick_child *Ctmp;
		newick_child *Dtmp;
		
		float tmp;
		tmp=(root->dist);
		(root->dist)=(rerootbranch->longestbranch->node->dist);
		(rerootbranch->longestbranch->node->dist)=tmp;
		(root->childNum)--;
		(rerootbranch->longestbranch->node->childNum)++;
		
		Atmp=root;
		Ctmp=rerootbranch->longestbranch->next;
		Dtmp=rerootbranch->longestbranch->node->child;
		
		root=rerootbranch->longestbranch->node;
		newn->node=Atmp;
		newn->next=Dtmp;
		root->child=newn;
		
		(rerootbranch->correspondingChild)->next=Ctmp;
		
		root->parent=NULL;
		
		
	}
	
	return root;
}


rerootReturn* findLongestBranch(newick_node *root, rerootReturn *rerootbranch)
{	
	newick_child *child;
	newick_child *childlast=NULL;
	float branchDistance;
	float longestdistance=0;
	int possible=0;
	
	if (root->childNum > 1)
	{
		child = root->child;
		
		while (child != NULL)
		{
			if (child->node->childNum>-1)
			{
				possible++;
				
				branchDistance=(child->node->dist) + (child->node->totaldistance);
				
				if (branchDistance >= longestdistance)
				{
					longestdistance=branchDistance;
					if (rerootbranch == NULL) {
						rerootbranch = malloc (sizeof *rerootbranch);
						if (rerootbranch==NULL)
					    {
							fprintf(stderr, "Malloc Problem Fix Me\n");
							exit(9);
					    }	
					}
					
					rerootbranch->longestbranch=child;
					if (childlast != NULL)
					{
						rerootbranch->correspondingChild=childlast;
					}
					else
					{
						rerootbranch->correspondingChild=NULL;
					}
					
				}
			}
			
			childlast = child;
			child = child->next;
		}
	}
	
	if (possible==0)
	{
		free(rerootbranch);
		rerootbranch=NULL;
	}
	return rerootbranch;
	
}



int checknumberofroots(newick_node *root, int numroots, int forPrint, int time, FILE *j)
{
	int numrootsreturn=0;
	newick_child *child;
	
	child = root->child;

	while (child != NULL)
	{
		numrootsreturn=checknumberofroots(child->node,0, forPrint, time, j);
		child=child->next;
		numroots=numroots+numrootsreturn;
		forPrint=numroots+forPrint;
	}
	
	
	if (root->roothere==1)
	{

		
		int ElementsPresent;
		if (time==2)
		{
			ElementsPresent=printTaxonNew(root,0,forPrint, root->timepoint,j);
			fprintf(j,"are in a cluster\n");
		}
		

		if (ElementsPresent==1)
		{
			numroots++;
			forPrint++;
		}
	}
	
	return numroots;
}


int checkpositionofroots(newick_node *root, int correct, float L)
{
	if (correct==1)
	{
		if (root->childNum==0)
		{
			if (root->roothere==1)
			{
				root->sequence=0;
			}
			else
			{
				root->sequence=1;
			}
		}
		else
		{
			root->sequence=0;
			
			newick_child*child;
			
			child=root->child;
			
			while (child!=NULL)
			{
				correct=checkpositionofroots(child->node,correct, L);
				
				if (child->node->roothere==1 && rint(child->node->dist*L)<30)
				{
					root->rootahead=1;
				}
				else
				{
					if (child->node->sequence==1 && rint(child->node->dist*L)<30)
					{
						root->sequence=1;		
					}
					
					if (child->node->rootahead==1 && rint(child->node->dist*L)<30)
					{
						root->rootahead=1;
					}
				}
				
				child=child->next;
			}
			
			if (root->roothere!=1)
			{
				if (root->parent!=NULL)
				{
					if (root->parent->roothere==1 && root->sequence==1 && root->rootahead==1)
					{

						if (rint(root->dist*L)<30 && root->numElements>0)
						{

							root->roothere=1;
							root->sequence=0;
							root->parent->sequence=0;
							if (root->parent!=NULL)
							{
								if (root->numElements==1)
								{
									root->parent->numElements=root->parent->numElements-root->numElements;
								}
								if (root->parent->numElements<=0)
								{
									root->parent->roothere=0;

								}
								
							}
							correct=checkpositionofroots(root,correct, L);
						}
					}
					
				}
				else
				{

					root->numrootsahead=0;
					root->sequence=0;
					
					child=root->child;
					while (child!=NULL)
					{						
						if (child->node->roothere==1)
						{
							root->numrootsahead++;
						}
						if (child->node->sequence==1)
						{
							root->sequence=1;
						}
						
						child=child->next;
					}
					if (root->sequence==1 && root->numrootsahead>1 && root->numElements>0)
					{
						
						root->roothere=1;
						root->sequence=0;
						correct=checkpositionofroots(root,correct, L);
						
					}
					else
					{
						root->roothere=0;
					}

				}
			}
			
			
		}
	}
	else
	{
		correct=0;
	}
	return correct;
}



