#ifndef __NEWICKFORM_H__
#define __NEWICKFORM_H__

typedef struct newick_child
{
	struct newick_node *node;
	struct newick_child *next;
} newick_child;

typedef struct newick_node
{
	char *taxon;
	char *seq;
	float dist;
	int unknownmarker;
	int recombinant;
	int childNum;
	int numElements;
	int cluster;
	struct newick_child *child;
	struct newick_node *parent;
	float longestbranch;
	float totaldistance;
	int printed;
	int roothere;
	int possibleroothere;
	int clusterless;
	int sequence;
	int rootahead;
	int numrootsahead;
	int clusternumbers[5];
	int combinedroothere;
	int nodebreak;
	int endingcluster;
	int timepoint;
} newick_node;


typedef struct rerootReturn
	{
		struct newick_child *longestbranch;
		struct newick_child *correspondingChild;
	} rerootReturn;

typedef struct fReturn
	{
		int *farray;
		int TFBranch;
	} fReturn;

short sDEBUG;


extern int CountClusters(newick_node *root, fReturn *farray, int Threshold, int L, int maxMutation, int timepoint, int timepointcluster);
extern void FindNumElements(newick_node *root);
extern int AverageBranch(newick_node *root);
extern int CutAverageBranch(newick_node *root, int maxMutation, fReturn *farray, float L, int timepoint, int timepointcluster);
extern int printTaxon(newick_node *root, int timepoint, int timepointcluster);
extern newick_node* parseTree(const char *const str);
extern void printTree(newick_node *root, FILE *g);
extern void printRepeatTree(struct newick_node* root, FILE* g, int numbranch, int* multiple, char ** probarray);
extern newick_node* changeRoot(newick_node *root, newick_child *newn, rerootReturn* rerootbranch);
extern rerootReturn* findLongestBranch(newick_node *root, rerootReturn *rerootbranch);
extern void printNodes(const newick_node *const root);
extern int checknumberofroots(newick_node *root, int numroots, int forPrint, int time, FILE *j);
extern int checkpositionofroots(newick_node *root, int correct, float L);
extern newick_node* analyzeTimePoint (char *pcInputFile, int timepoint, FILE *j);
extern newick_node* analyzeCombinedTree(char *pcInputFile, int timepoint);
extern void combinetimepoints(newick_node *root);
extern int printFinalClustering (newick_node *root, int cluster, int present, FILE *j);
extern int combinenodes(newick_node *root, int cluster);
extern int convertClustering(newick_node *root, newick_node *time0, newick_node *time1, newick_node *time2, newick_node *time3, newick_node *time4, int notcomplete);
extern void makePrintedZero(newick_node * root);


#endif

